/*
 * ApproximationBuilder.cpp
 *
 *  Created on: Sep 17, 2009
 *      Author: testylie
 */

#include "ApproximationBuilder.h"
#include <PolynomialSystem.h>
//#include "CurvatureTensor.h"
#include <math.h>
#include <nlopt.hpp>

//#ifndef GMP_RAT

using namespace std;
using namespace arma;

namespace nltool {

ApproximationBuilder::ApproximationBuilder() :
		nonlinearSystem(NULL), minCurvature(1e-4), dimension(0) {
}

ApproximationBuilder::ApproximationBuilder(int dimension) :
		dimension(dimension), nonlinearSystem(NULL), useTransformation(true), minCurvature(
				1e-4), curvatureEstimationCoef(10), useDisplacement(true) {
	// variables initialisation
	maxCurvature = -1;
	curvatureEstimationDomain = Box(stdVector(dimension, -5),
			stdVector(dimension, 5));
}

ApproximationBuilder::ApproximationBuilder(int dimension, DynamicalSystem* sys,
		bool useT) :
		dimension(dimension), nonlinearSystem(sys), useTransformation(useT), minCurvature(
				1e-4), curvatureEstimationCoef(10), useDisplacement(true) {
	// variables initialisation
	maxCurvature = -1;
	curvatureEstimationDomain = Box(stdVector(dimension, -5),
			stdVector(dimension, 5));
}

ApproximationBuilder::~ApproximationBuilder() {
//	if (nonlinearSystem != NULL)
//		delete nonlinearSystem;
}

/*
 * ##########################################
 * TRANSFORMATION MATRIX COMPUTATION
 * ##########################################
 */

// TEST FUNCTION
void ApproximationBuilder::testCurvature(Matrix& curvature, int nbTest) {

	int nbdirection = 5;
	int nbTestReal = 0;
	int nbTestOk = 0;
	Number maxViolation = 0;

	while (nbTestReal < (nbTest * nbdirection)) {
		vector<double> pt(dimension);
		// now we generate a random point
//		for (unsigned i = 0; i < dimension; i++) {
//			pt[i] = ((rand() % 100 + 1) / 100.0)
//					* (domain.getBu()[i] - domain.getBl()[i])
//					+ domain.getBl()[i];
//		}

// now we create random direction
		for (int i = 0; i < nbdirection; i++) {
			Vector dir(dimension);
			for (unsigned j = 0; j < dimension; j++) {
				dir[j] = (rand() % 10 + 1) / 5.0 - 1.0;
			}
			// we normalize this direction
			Number norm = arma::norm(dir, 2);

			if (norm != 0) {
				dir = dir / norm;
				cout << "direction: " << dir;
				Matrix tmp = dir.t() * curvature * dir;
				cout << "tmp: " << tmp << endl;
				//				Matrix tmp2 = tmp*dir;
				Number dtQd = tmp[0];
				for (unsigned z = 0; z < dimension; z++) {

					//double dHdt = fCurv(z,pt,dir.data,dir.data, fData);
					tmp = dir.t() * cstHessian[z] * dir;
					Number dtHd = tmp[0];

					nbTestReal++;
					if (fabs(dtQd) >= fabs(dtHd)) {
						nbTestOk++;
					} else {
						Number violation = fabs(dtHd) - fabs(dtQd);
						cout << "Warning: curvature tensor failed :"
								<< violation << endl;
						if (maxViolation < violation)
							maxViolation = violation;
					}
					cout << endl;
				}
			}
		}

	}
	cout << nbTestOk << " test succeed on " << nbTestReal << endl;
	cout << "maxViolation: " << maxViolation << endl;
	wait();
}

Matrix ApproximationBuilder::computeTransformationMatrix() {
	Matrix res;

	if (nonlinearSystem->getSystemType() != QUADRATIC) {
		assert(curvatureEstimationDomain.getDimension() > 0);
	}

	cstHessian = computeHessianMatrix();

	//cout << "Hessian matrix computed: " << endl;
	//for (unsigned i = 0; i < cstHessian.size(); i++)
	//	cout << cstHessian[i] << endl;

	// compute curvature matrix
	curvature = computeCurvatureMatrix(cstHessian);

	// compute transformation matrix T
	transformation = computeTransformationMatrix(curvature);

	return res;
}

Matrix ApproximationBuilder::computeTransformationMatrix(vector<Matrix> &hess) {
	Matrix res;
	// compute curvature tensor matrix
	curvature = computeCurvatureMatrix(hess);

	// compute transformation matrix T
	transformation = computeTransformationMatrix(curvature);

	return res;
}

Matrix ApproximationBuilder::computeTransformationMatrix(Matrix &curvature) {
	Matrix res(zeros(dimension, dimension));
	Matrix eigenvec;
	Vector eigenVal;

	arma::eig_sym(eigenVal, eigenvec, curvature);
	for (unsigned i = 0; i < dimension; i++)
		res(i, i) = Number(sqrt(fabs(eigenVal(i)) / maxCurvature));
	res = eigenvec * res * eigenvec.t();
	return res;
}

Matrix ApproximationBuilder::computeCurvatureMatrix(vector<Matrix>& hessian) {

	// compute curvature matrix from eigen decomposition of hessian
	Matrix result = zeros(dimension, dimension);
	maxCurvature = -1;
	Vector maxEigenVal(dimension);
	Matrix maxEigenVec;
//	int maxHessian = 0;
//	int indMax = 0;
	Number secondMax = 0;

	// test solution

	for (unsigned i = 0; i < dimension; i++) {
		// get the absolute value of the eigenvalues
		Matrix eigenvec;
		Vector eigenVal;
		arma::eig_sym(eigenVal, eigenvec, hessian[i]);
		eigenVal = sqrt(square(eigenVal));

		// find the maximal eigen Value
		Number maximalEigenValue = max(eigenVal);
		if (maxCurvature < maximalEigenValue) {
			// check the presence of second maximal eigen value
			for (unsigned j = 0; j < dimension; j++) {
				if ((eigenVal[j] < maximalEigenValue)
						&& (eigenVal[j] > secondMax))
					secondMax = eigenVal[j];
			}
			if (secondMax < maxCurvature)
				secondMax = maxCurvature;

			maxCurvature = maximalEigenValue;
			maxEigenVec = eigenvec;
			maxEigenVal = eigenVal;
		} else {
			if (secondMax < maximalEigenValue)
				secondMax = maximalEigenValue;
		}

	}

	if (closeToZero(maxCurvature))
		result = (eye(dimension, dimension));
	else {
		if (closeToZero(secondMax))
			secondMax = maxCurvature / 10.0;
		result = eye(dimension, dimension) * maxCurvature;
		for (unsigned i = 0; i < maxEigenVal.size(); i++) {
			if (maxEigenVal[i] < maxCurvature)
				result(i, i) = secondMax;
		}
		result = maxEigenVec * result * maxEigenVec.t();

//	testCurvature(curvature, 1000);
	}
	return result;
}

vector<Matrix> ApproximationBuilder::computeHessianMatrix() {
	//cout<<"computing hessian with dimension: "<<dimension<<endl;
	std::vector<Matrix> res(dimension);
	Vector pt(dimension);
	Hpolyhedron domain;

	switch (nonlinearSystem->getSystemType()) {
	case QUADRATIC:
		for (unsigned d = 0; d < dimension; d++) {
			Matrix tmp(dimension, dimension);
			for (unsigned i = 0; i < dimension; i++) {
				for (unsigned j = i; j < dimension; j++) {
					// double derivative function (only constant for the moment)
					DynamicalSystem& deriv1 = nonlinearSystem->derivate(i);
					DynamicalSystem& deriv2 = deriv1.derivate(j);
					tmp(i, j) = (deriv2.eval(pt))[d];
					tmp(j, i) = tmp(i, j);
					delete &deriv1;
					delete &deriv2;
				}
			}
			res[d] = tmp;
		}
		break;

	case POLYNOMIAL:
		assert(curvatureEstimationDomain.getDimension() > 0);
		domain = Hpolyhedron(curvatureEstimationDomain);
		for (unsigned d = 0; d < dimension; d++) {
			Matrix tmp(dimension, dimension);
			for (unsigned i = 0; i < dimension; i++) {
				for (unsigned j = i; j < dimension; j++) {
					// double derivative function (only covalMaxnstant for the moment)
					DynamicalSystem& deriv1 = nonlinearSystem->derivate(i);
					DynamicalSystem& deriv2 = deriv1.derivate(j);
					Number max =
							static_cast<PolynomialSystem&>(deriv2).maximize(d,
									domain.getTemplates(),
									domain.getCoefficients(), pt);
					Number min =
							static_cast<PolynomialSystem&>(deriv2).minimize(d,
									domain.getTemplates(),
									domain.getCoefficients(), pt);
					if (fabs(max) > fabs(min))
						tmp(i, j) = max;
					else
						tmp(i, j) = min;
					tmp(j, i) = tmp(i, j);
					delete &deriv1;
					delete &deriv2;
				}
			}
			res[d] = tmp;
		}
		break;

	case POINTER:
		assert(curvatureEstimationDomain.getDimension() > 0);
		for (unsigned d = 0; d < dimension; d++) {
			Matrix tmp(dimension, dimension);
			for (unsigned i = 0; i < dimension; i++) {
				for (unsigned j = i; j < dimension; j++) {
					// double derivative function (only covalMaxnstant for the moment)
					tmp(i, j) =
							static_cast<PointerSystem&>(*nonlinearSystem).getMaxCurvature(
									d, i, j, curvatureEstimationDomain);
					tmp(j, i) = tmp(i, j);
				}
			}
			res[d] = tmp;
		}
		break;

	default:
		break;
	}

	return res;
}

Number ApproximationBuilder::computeMaximalCurvature() {
	// compute constant hessian matrix
	cstHessian = computeHessianMatrix();

	// compute curvature matrix
	curvature = computeCurvatureMatrix(cstHessian);

	return maxCurvature;
}

void ApproximationBuilder::computeCurvatureDomain(Hpolyhedron set) {
	Box bbox(set);
	stdVector bl = bbox.getBl();
	stdVector bu = bbox.getBu();
	stdVector interval = bu - bl;
	Number maxInterval = max(interval);
	for (unsigned i = 0; i < bbox.getDimension(); i++) {
		bl[i] -= curvatureEstimationCoef * maxInterval;
		bu[i] += curvatureEstimationCoef * maxInterval;
	}
	curvatureEstimationDomain = Box(bl, bu);
}

/*
 * ##########################################
 * SIMPLEX CREATION
 * ##########################################
 */

Simplex ApproximationBuilder::createSimplex(Hpolyhedron initialSet,
		Number error) {

	// check if we need to compute the curvature tensor or the transformation matrix
	if (maxCurvature < 0) {
		if (nonlinearSystem->getSystemType() != QUADRATIC) {
			// compute a curvature estimation domain
			computeCurvatureDomain(initialSet);
		}
		if (useTransformation)
			computeTransformationMatrix();
		else
			computeMaximalCurvature();
	}


	bool succeed = false;
	bool split_required = false;

	Box bbox(initialSet);
	Vertices vertices = bbox.generateVertices();
	Vector direction = computeDirectionalVector(initialSet);

	Simplex domain;
	do {
		//cout<<"direction: "<<direction.t()<<endl;
		Number rmcd = sqrt(2 * error / maxCurvature);

		//##############################
		// STEP 1: transformation matrix

		// transform initialSet and direction
		Vertices tmpVertices = vertices;
		if (useTransformation) {
			assert(!transformation.is_empty());
			direction = transformation * direction;
			for (unsigned i = 0; i < tmpVertices.size(); i++)
				tmpVertices[i] = transformation * tmpVertices[i];
		}
		Box bbox2 = Box(dimension, tmpVertices);
		Hpolyhedron tmpPoly(bbox2);

		//##############################
		// STEP 2: create equilateral simplex using vertex representation

		//if (initialSimplex.getVertices().size() == 0)
		initialSimplex = createRegularSimplex(rmcd, dimension);

		// align simplex
		Vertices simplexVertices(initialSimplex.getVertices());
		alignToDirection(simplexVertices, direction);

		// place the simplex in space
		split_required = !placeSimplex(simplexVertices, tmpPoly,
				bbox2.getMaxInterval());
//		cout << "split required: " << split_required << endl;
		if (split_required) {
			cout << "split required!" << endl;
			domain = Simplex();
			break;
		}

		// apply affine transformation with T
		// apply transformation on each vertices of the polyhedron

		if (useTransformation) {
			Matrix inverseTransformation = inv(transformation);
			for (unsigned int i = 0; i < simplexVertices.size(); i++)
				simplexVertices[i] = inverseTransformation * simplexVertices[i];
		}

		domain = Simplex(dimension, simplexVertices);
		if (nonlinearSystem->getSystemType() != QUADRATIC) {
			// check its inclusion in curvature estimation domain
			Hpolyhedron polyDomain(domain);
			if (!curvatureEstimationDomain.include(polyDomain)) {
				curvatureEstimationCoef *= 2;
				computeCurvatureDomain(initialSet);
				if (useTransformation)
					computeTransformationMatrix();
			} else
				succeed = true;
		} else
			succeed = true;
	} while (!succeed);

	return domain;
}

Simplex ApproximationBuilder::createRegularSimplex(Number radius,
		int dimension) {
	// let origin be the center of the sphere
	Vertices points(dimension + 1, zeros(dimension));

	// compute first radius and first sphere center
	Number current_radius = radius;
	Vector current_center(zeros(dimension));
	// initial sphere center = origin

	// recursive part
	for (int i = 0; i < dimension; i++) {
		// new point on
		for (int j = 0; j < i; j++) {
			//current_point[i][j]= dimensionAlreadySet[j];
			points[i][j] = current_center[j];

		}
		points[i][i] = current_center[i] + current_radius;
		for (int j = i + 1; j < dimension; j++) {
			//other coordinate set to 0
			points[i][j] = 0;
		}

		// now we compute the hyper plan
		if ((dimension - i) == 1) {
			// we compute the last vertex
			points[dimension][dimension - 1] = current_center[dimension - 1]
					- current_radius;
			for (int j = 0; j < (dimension - 1); j++) {
				points[dimension][j] = current_center[j];
			}
		} else {
			// compute the new sphere center
			Number dFromCenter = (dimension - i + 1) * current_radius
					/ (dimension - i); //not safe over 3d
			current_center[i] = current_center[i] + current_radius
					- dFromCenter;

			// compute the new radius
			Number h = current_radius;
			Number o = current_radius / (dimension - i);
			current_radius = sqrt(fabs(h * h - o * o));
		}
	}

	return Simplex(dimension, points);
}

Vector ApproximationBuilder::computeDirectionalVector(Hpolyhedron &initial) {

	Box bbox(initial);
	Vector direction = nonlinearSystem->eval(bbox.centroid());
	direction /= arma::norm(direction, 2);
	return -direction;
}

void ApproximationBuilder::alignToDirection(Vertices& points,
		Vector direction) {
	vector<Matrix> rotations(dimension * 2, zeros(dimension, dimension));
	int nbM = 0;
	int dim = dimension;

	for (int i = dim - 1; i > 0; i--) {
		// we compute the angle
		double angle = 0;
		angle = atan2(direction[i], direction[i - 1]);

		// now we create the rotation matrix
		for (int j = 0; j < dim; j++) {
			rotations[nbM](j, j) = 1;
		}
		rotations[nbM](i - 1, i - 1) = cos(angle);
		rotations[nbM](i - 1, i) = -sin(angle);
		rotations[nbM](i, i - 1) = sin(angle);
		rotations[nbM](i, i) = cos(angle);

		// we apply the matrix to the direction vector
		direction = rotations[nbM] * direction;

		// we inverse (=transpose with rotation matrix) this matrix
		rotations[nbM] = rotations[nbM].t();
		nbM++;
	}

	// now lets use all this rotation matrix
	for (int i = (nbM - 1); i >= 0; i--) {
		for (unsigned int j = 0; j < points.size(); j++) {
			points[j] = rotations[i].t() * points[j];
		}
	}
}

bool ApproximationBuilder::placeSimplex(Vertices& points, Hpolyhedron& init,
		Number setMaxSize) {

	Vector centroidInit = Box(init).centroid();
	Vector centroidSimplex = computeCentroid(points);
	Vector alignVector(dimension);
	bool included = false;

	Vector translation(zeros(dimension));
	Box bboxSimplex(dimension, points);
	Number proportion = setMaxSize / bboxSimplex.getMaxInterval();
	alignVector = centroidInit - centroidSimplex;

	for (unsigned int i = 0; i < points.size(); i++)
		points[i] += alignVector;

	if (useDisplacement) {
		Vertices tmpPoints = points;
		if (proportion < 0.5)
			translation = (-1 * proportion + 0.5) * (centroidInit - points[0]);
		// place the simplex
		for (unsigned int i = 0; i < points.size(); i++)
			for (unsigned j = 0; j < dimension; j++)
				tmpPoints[i][j] += translation[j];

		Simplex tmpSimplex(dimension, tmpPoints);
		Hpolyhedron tmpPoly(tmpSimplex);
		if (tmpPoly.includes(init)) {
			included = true;
		} else {
			// if no direct inclusion we try some other placement
			while ((!included) && (proportion < 0.5)) {
				tmpPoints = points;
				proportion *= 2;
				translation = zeros(dimension);

				if (proportion < 0.5)
					translation = (-1 * proportion + 0.5)
							* (centroidInit - points[0]);
				else
					cout << "proportion failed" << endl;
				//cout << "translation: " << translation << endl;
				// place the simplex
				for (unsigned i = 0; i < points.size(); i++)
					for (unsigned j = 0; j < dimension; j++)
						tmpPoints[i][j] += translation[j];
				tmpSimplex = Simplex(dimension, tmpPoints);
				tmpPoly = Hpolyhedron(tmpSimplex);
				if (tmpPoly.includes(init)) {
					included = true;
				} else
					proportion *= 2;
			}
		}
		points = tmpPoints;
		return included;
	} else {
		Simplex tmpSimplex(dimension, points);
		Hpolyhedron tmpPoly(tmpSimplex);
		return tmpPoly.includes(init);
	}
}

LinearSystem ApproximationBuilder::createApproximation(Hpolyhedron& initialSet,
		Hpolyhedron& approximationDomain, Number errorBound,
		bool& requiresSplitting) {

	requiresSplitting = false;

	// phase 1: create the approximation domain
	Simplex approximationSimplex(createSimplex(initialSet, errorBound));
	// phase 2: compute the affine interpolation
	Matrix A;
	Vector b;

	if (!(approximationSimplex.getDimension() > 0)) {
		requiresSplitting = true;
		return LinearSystem(0, A, b);
	}

	approximationSimplex.linearInterpolation(*nonlinearSystem, A, b);

	// phase 3: create the final approximate system and approximation domain
	// TODO: remove this
//	errorBound = 0;

	stdVector boundU(dimension, errorBound);
	stdVector boundL(dimension, -errorBound);
	Box inputSet(boundL, boundU);
	Set* setPtr = new Box(boundL, boundU);
	LinearSystem approximation(dimension, A, b, setPtr);
	approximationDomain = Hpolyhedron(approximationSimplex);

//	cout << "initial input set: " << inputSet << endl;

	Set* newInputSet = new Box(
			estimateApproximationerror(nonlinearSystem,
					static_cast<DynamicalSystem*>(&approximation),
					approximationDomain,errorBound));
	approximation.setInputSet(newInputSet);

	cout << "new input set: " << static_cast<Box&>(*newInputSet) << endl;

	//delete setPtr;
	return approximation;
}

Vector ApproximationBuilder::computeCentroid(Vertices& points) {
	Vector center(zeros(dimension));
	for (unsigned i = 0; i < dimension; i++) {
		for (unsigned int j = 0; j < points.size(); j++) {
			center[i] = center[i] + points[j][i];
		}
	}
	for (unsigned i = 0; i < dimension; i++) {
		center[i] = center[i] / points.size();
	}
	return center;
}

Matrix ApproximationBuilder::getTransformation() const {
	return transformation;
}

bool ApproximationBuilder::isUseTransformation() const {
	return useTransformation;
}

void ApproximationBuilder::setTransformation(Matrix transformation) {
	this->transformation = transformation;
}

void ApproximationBuilder::setMaxCurvature(double maxCurvature) {
	this->maxCurvature = maxCurvature;
}

void ApproximationBuilder::setUseTransformation(bool useTransformation) {
	this->useTransformation = useTransformation;
}

// ###############################
// Error Approximation computation
// ###############################
using namespace nlopt;

double approximationErrorFunc(const std::vector<double> &x,
		std::vector<double> &grad, void* f_data) {
	Vector pt = arma::conv_to<Vector>::from(x);
	ErrorEstimationParameters* param =
			static_cast<ErrorEstimationParameters*>(f_data);
	return fabs(
			(param->originalSystem->eval(pt))[param->indice]
					- (param->approximateSystem->eval(pt))[param->indice]);
}

Box estimateApproximationerror(DynamicalSystem* original,
		DynamicalSystem* approximation, Hpolyhedron& domain, Number initialerror) {
//	using namespace nlopt;
	ErrorEstimationParameters param;
	param.originalSystem = original;
	param.approximateSystem = approximation;
	param.indice = 0;
	int dimension = original->getDimension();
	Box newDomain(domain);
	stdVector bl(dimension);
	stdVector bu(dimension);

	// create the root object
	opt my_opt(LN_COBYLA, dimension);

	// set objective function

	// add bounds
	my_opt.set_lower_bounds(newDomain.getBl());
	my_opt.set_upper_bounds(newDomain.getBu());

	// termination condition
	my_opt.set_xtol_abs(1.e-6);
	//my_opt.set_ftol_rel(1.e-6);
	my_opt.set_maxtime(0.1);

	// initial guess
	std::vector<double> x(dimension, 0);
	for (unsigned i = 0; i < dimension; i++)
		x[i] = newDomain.getBl()[i]
				+ (newDomain.getBu()[i] - newDomain.getBl()[i]) / 2;

	for (unsigned i = 0; i < dimension; i++) {
		param.indice = i;
		double max, min = 0;
		my_opt.set_max_objective(approximationErrorFunc, (void*) &param);
		result res = my_opt.optimize(x, max);
		if (!res)
			throw(runtime_error("Error in NLopt: " + res));
		if(max>initialerror) max = initialerror;
		bu[i] = max;

		my_opt.set_min_objective(approximationErrorFunc, (void*) &param);
		res = my_opt.optimize(x, min);
		if (!res)
			throw(runtime_error("Error in NLopt: " + res));
		if(min<-initialerror) min = -initialerror;
		bl[i] = min;
	}

//	my_opt.set_min_objective(approximationErrorFunc, (void*) this);
//	res = my_opt.optimize(x, min);
//	if (!res)
//		throw(runtime_error("Error in NLopt: " + res));

	return Box(bl, bu);
}

} // namespace nltool

/*
 * ##########################################
 * UTILS
 * ##########################################
 */
/*
 void ApproximationBuilder::translatePoly(ReachableSet &poly, stdVector t){
 stdMatrix points = poly.vertices();
 for(int i=0;i<poly.nb_vertices();i++){
 for(int j=0;j<poly.dim();j++){
 points[i][j] += t[j];
 }
 }
 poly.Vrep_make(poly.dim(),points);
 }

 void ApproximationBuilder::applyTransformation(stdMatrix &T, ReachableSet &poly){
 stdMatrix points = poly.vertices();
 //cout << "points = " << points << "\n";

 //cout << "T = " << T << "\n";

 for(int i=0;i<poly.nb_vertices();i++){
 points[i] = T*points[i];
 }

 //cout << "New points = " << points << "\n";
 poly.Vrep_make(poly.dim(),points);
 }*/

/*
 * ####################################
 *	Is inside
 * ####################################
 */

/*
 void ApproximationBuilder::setDomain(stdVector bl, stdVector bu){
 this->bl = bl;
 this->bu = bu;
 }
 */

