/*
 * Box.cpp
 *
 *  Created on: Nov 20, 2012
 *      Author: romain
 */

#include "Box.h"
#include "Hpolyhedron.h"

namespace nltool {

Box::Box() {
	dimension = 0;
}

Box::~Box() {
	// TODO Auto-generated destructor stub
}

Box::Box(const stdVector& l, const stdVector& u) :
		bl(l), bu(u) {
	assert(l.size() == u.size());
	dimension = l.size();
}

Box::Box(unsigned dim, const Vertices& vertices) {
	bl = stdVector(dim);
	bu = stdVector(dim);
	dimension = dim;
	double l, u;
	for (unsigned i = 0; i < dim; i++) {
		l = vertices[0][i];
		u = vertices[0][i];
		for (unsigned int j = 1; j < vertices.size(); j++) {
			if (l > vertices[j][i])
				l = vertices[j][i];
			if (u < vertices[j][i])
				u = vertices[j][i];
		}
		bl[i] = l;
		bu[i] = u;
	}
}

Box::Box(unsigned dim, const Template& A, const Vector& b) {
	dimension = dim;
	bl = stdVector(dimension, 0);
	bu = stdVector(dimension, 0);

	for (unsigned int i = 0; i < dimension; i++) {
		Vector obj = arma::zeros(dimension);
		obj[i] = -1;
		bl[i] = -simplex(obj, A, b);
		obj[i] = 1;
		bu[i] = simplex(obj, A, b);
	}
}

Box::Box(Hpolyhedron& poly) {
	dimension = poly.getDimension();
	Box tmpBox = poly.getBbox();
	if (tmpBox.getDimension() == 0) {

		Template A(poly.getTemplates());
		Vector b(poly.getCoefficients());
		dimension = poly.getDimension();
		bl = stdVector(dimension, 0);
		bu = stdVector(dimension, 0);

		for (unsigned int i = 0; i < dimension; i++) {
			Vector obj(arma::zeros(dimension));
			obj[i] = -1;
			bl[i] = -simplex(obj, A, b);
			obj[i] = 1;
			bu[i] = simplex(obj, A, b);
		}
	} else {
		bl = tmpBox.getBl();
		bu = tmpBox.getBu();
	}
}

std::vector<int> int2binary(int N, int num) {
	std::vector<int> bin(N);
	for (int j = 0; j < N; j++) {
		int k = 1 << j;
		bin[j] = 0;
		if (num & k)
			bin[j] = 1;
	}
	return (bin);
}

Vertices Box::generateVertices() {
	Vertices vertices;
	std::vector<int> offset(dimension);
	Vector x(dimension);
	stdVector len(dimension);
	for (unsigned i = 0; i < dimension; i++)
		len[i] = bu[i] - bl[i];
	for (int k = 0; k < (1 << dimension); k++) {
		offset = int2binary(dimension, k);
		for (unsigned i = 0; i < dimension; i++) {
			if (offset[i] == 1)
				x[i] = bl[i] + len[i];
			else
				x[i] = bl[i];
		}
		vertices.push_back(x);
	}
	return vertices;
}

std::vector<Polynom> Box::computeUnitBoxMap() {
	vector<Polynom> result;
	for (unsigned i = 0; i < dimension; i++) {
		Number tmp = bu[i] - bl[i];
		multiIndice tmpMultiIndice(dimension, 0);
		vector<Monom> tmpMonoms;
		tmpMultiIndice[i] = 1;
		tmpMonoms.push_back(Monom(dimension, tmp, tmpMultiIndice));
		tmpMultiIndice[i] = 0;
		tmpMonoms.push_back(Monom(dimension, bl[i], tmpMultiIndice));
		result.push_back(Polynom(dimension, tmpMonoms));
	}
	return result;
}

Vector Box::centroid() {
	Vector result(dimension);
	for (unsigned int i = 0; i < dimension; i++)
		result[i] = bl[i] + ((bu[i] - bl[i]) / 2);
	return result;
}

bool Box::include(Hpolyhedron& poly) {
	for (unsigned i = 0; i < dimension; i++) {
		Vector obj(arma::zeros(dimension));
		obj[i] = 1;
		if ((simplex(obj, poly.getTemplates(), poly.getCoefficients()) > bu[i])
				|| (simplex(obj, poly.getTemplates(), poly.getCoefficients(),
						false) < bl[i]))
			return false;
	}
	return true;
}

Number Box::getMaxInterval() {
	Number result = bu[0] - bl[0];
	for (unsigned i = 0; i < dimension; i++)
		if (result < bu[i] - bl[i])
			result = bu[i] - bl[i];
	return result;
}

Number Box::getMinInterval() {
	Number result = bu[0] - bl[0];
	for (unsigned i = 0; i < dimension; i++)
		if (result > bu[i] - bl[i])
			result = bu[i] - bl[i];
	return result;
}

stdVector Box::getBl() const {
	return bl;
}

stdVector Box::getBu() const {
	return bu;
}

void Box::setBl(stdVector bl) {
	this->bl = bl;
}

Number Box::optimize(const Vector& direction, Vector& optPoint, bool max) {
	optPoint = arma::zeros(dimension);
	for (unsigned i = 0; i < direction.n_elem; i++)
		if (direction[i] < 0)
			optPoint[i] = bl[i];
		else
			optPoint[i] = bu[i];
	return arma::dot(direction, optPoint);
}

void Box::setBu(stdVector bu) {
	this->bu = bu;
}

}

