/*
 * HybridizationInterface.cpp
 *
 *  Created on: Feb 8, 2013
 *      Author: romain
 */

#include "HybridizationInterface.h"
#include "ReachHybridization.h"

namespace nltool {

HybridizationInterface::HybridizationInterface() :
		dimension(0), systemPtr(NULL), systemReady(false), setReady(false) {
}

HybridizationInterface::HybridizationInterface(int dim) :
		dimension(dim), systemPtr(NULL), systemReady(false), setReady(false) {
}

HybridizationInterface::~HybridizationInterface() {
}

void HybridizationInterface::setSystemDynamic(unsigned dim,
		double (*f)(const int x, const std::vector<double>& pt, void* data),
		double (*ddf)(const int x, const std::vector<double> &pt, const int dx,
				const int dy, void* data), void* data) {
	if (systemPtr != NULL)
		delete systemPtr;
	dimension = dim;
	systemPtr = new PointerSystem(dim, f, NULL, ddf, data, false);
	systemReady = true;
}

void HybridizationInterface::setSystemDynamic(unsigned dim, std::string str) {
	if (systemPtr != NULL)
		delete systemPtr;
	dimension = dim;
	systemPtr = new PolynomialSystem(dim, str);
	systemReady = true;
}

void HybridizationInterface::setInitialSet(unsigned dim, unsigned nbConstraint,
		vector<vector<double> > &A, vector<double> &b) {
	assert(dimension == dim);
	Template constraintsA(nbConstraint, arma::zeros(dim));
	Vector constraintB(arma::zeros(nbConstraint));
	for (unsigned i = 0; i < nbConstraint; i++) {
		for (unsigned j = 0; j < dim; j++)
			constraintsA[i][j] = A[i][j];
		constraintB[i] = b[i];
	}
	initialSet = Hpolyhedron(dim, constraintsA, constraintB);
	setReady = true;
}

HybridizationResult HybridizationInterface::computeApproximation(
		const double precision, vector<vector<double> > &resultA,
		vector<double> &resultb, vector<vector<double> > &domainA,
		vector<double> &domainb) {

	if(!systemReady || !setReady)
		throw(runtime_error("dynamical system and initial set must be initialized "
				"before calling the computeApproximation methods"));

	Hpolyhedron domain;
	bool requireSplitting = false;

	ApproximationBuilder approximationBuilder(dimension, systemPtr, false);

	LinearSystem approximation = approximationBuilder.createApproximation(
			initialSet, domain, precision, requireSplitting);

	if (requireSplitting)
		return REQUIRESPLITTING;

	//test
	//checkErrorBound(precision,approximation,*systemPtr,domain);

	Template domainAtmp = domain.getTemplates();
	Vector domainbtmp = domain.getCoefficients();

	// result writing
	domainA = vector<vector<double> >(domainAtmp.size(),
			vector<double>(dimension));
	domainb = vector<double>(domainbtmp.size());

	resultA = vector<vector<double> >(dimension, vector<double>(dimension));
	resultb = vector<double>(dimension);

	for (unsigned i = 0; i < domainAtmp.size(); i++) {
		for (unsigned j = 0; j < dimension; j++)
			domainA[i][j] = domainAtmp[i][j];
		domainb[i] = domainbtmp[i];
	}

	for (unsigned i = 0; i < dimension; i++) {
		for (unsigned j = 0; j < dimension; j++)
			resultA[i][j] = approximation.getA()(i, j);
		resultb[i] = approximation.getB()[i];
	}
	return SUCCESS;
}

} /* namespace nltool */
