/*
 * PointerSystem.cpp
 *
 *  Created on: Jan 22, 2013
 *      Author: romain
 */

#include "PointerSystem.h"
#include <nlopt.hpp>


namespace nltool {


using namespace nlopt;

PointerSystem::PointerSystem(){}

PointerSystem::PointerSystem(int dim,
		Number (*f)(const int x, const stdVector &pt, void* data),
		Number (*df)(const int x, const stdVector &pt, const int dx, void* data),
		Number (*ddf)(const int x, const stdVector &pt, const int dx, const int dy, void* data),
		void* data, bool discretized) {
	dimension = dim;
	fPtr = f;
	dfPtr = df;
	ddfPtr = ddf;
	dataPtr = data;
	systemType = POINTER;
	if (discretized)
		systemTimeType = DISCRETE;
	else
		systemTimeType = CONTINUOUS;

}

Vector PointerSystem::eval(const Vector& x) {
	stdVector pt = arma::conv_to<stdVector>::from(x);
	Vector result(dimension);
	for (unsigned i = 0; i < dimension; i++) {
		result[i] = fPtr(i, pt, dataPtr);
	}
	return result;
}

Vector PointerSystem::evalCurvature(const Vector& x, int dx, int dy) {
	stdVector pt = arma::conv_to<stdVector>::from(x);
	Vector result(dimension);
	for (unsigned i = 0; i < dimension; i++) {
		result[i] = ddfPtr(i, pt, dx, dy, dataPtr);
	}
	return result;
}

PointerSystem::~PointerSystem() {
	// TODO Auto-generated destructor stub
}

double PointerSystem::CurvatureFunc(const std::vector<double> &x,
		std::vector<double> &grad, void* f_data) {
	PointerSystem *t = static_cast<PointerSystem*>(f_data);
	return t->ddfPtr(t->curvatureFunctionX, x, t->curvatureFunctionDx,
			t->curvatureFunctionDy, t->dataPtr);
}

Number PointerSystem::getMaxCurvature(const unsigned i, const unsigned dx,
		const unsigned dy, const Box& domain) {
	curvatureFunctionX = i;
	curvatureFunctionDx = dx;
	curvatureFunctionDy = dy;

	// create the root object

	opt my_opt(LN_COBYLA, dimension);

	// set objective function

	// add bounds
	my_opt.set_lower_bounds(domain.getBl());
	my_opt.set_upper_bounds(domain.getBu());

	// termination condition
	my_opt.set_xtol_abs(1.e-6);
	my_opt.set_ftol_rel(1.e-6);

	// initial guess
	std::vector<double> x(dimension, 0);
	for (unsigned i = 0; i < dimension; i++)
		x[i] = domain.getBl()[i] + (domain.getBu()[i] - domain.getBl()[i]) / 2;

	double max = 0;
	double min = 0;
	my_opt.set_max_objective(CurvatureFunc, (void*) this);
	result res = my_opt.optimize(x, max);
	if (!res)
		throw(runtime_error("Error in NLopt: " + res));

	my_opt.set_min_objective(CurvatureFunc, (void*) this);
	res = my_opt.optimize(x, min);
	if (!res)
		throw(runtime_error("Error in NLopt: " + res));

	if(fabs(min)>fabs(max))
		return min;
	else
		return max;

	return 0;
}


}  // namespace nltool
