/**
 *
 * HybridizationSimplex is a class which creates a simplex
 * fitted to a set represented by a polyhedra linked to
 * a dynamical system
 *
 * @author Romain Testylier
 *
 */

#ifndef HYBRIDIZATIONSIMPLEX_H_
#define HYBRIDIZATIONSIMPLEX_H_

#include "utils.h"
#include "DynamicalSystem.h"
#include "LinearSystem.h"
#include "PointerSystem.h"
#include "PolynomialSystem.h"
#include "Hpolyhedron.h"
#include "Box.h"
//#include "CurvatureTensor.h"

namespace nltool {

using namespace std;

enum system_type {
	NOT_DEFINED = 1, HESSIAN_CONSTANT = 2, HESSIAN_NOT_CONSTANT = 3
};

class ApproximationBuilder {
public:

	ApproximationBuilder();

	/**
	 * Builder
	 * create the factory
	 *
	 * @param dimension system dimension
	 */
	ApproximationBuilder(int dimension);

	/**
	 * Builder
	 * create the factory
	 *
	 * @param dimension system dimension
	 * @param sys a non linear system
	 * @param useT Flag to use isotropic transformation matrix
	 */
	ApproximationBuilder(int dimension, DynamicalSystem* sys,
			bool useT = false);

	/** Destructor	*/
	virtual ~ApproximationBuilder();

	/** Function that creates an approximation of a nonlinear system by a
	 * linear system with input given a desired approximation error bound
	 *
	 * @param initialSet the current reachable set
	 * @param approximationDomain the approximation domain returned
	 * @param errorBound the precision paramater
	 * @return a linear system with input which approximate accurately
	 * the original nonlinear system in the approximation domain
	 */
	LinearSystem createApproximation(Hpolyhedron& initialSet,
			Hpolyhedron& approximationDomain, Number errorBound,
			bool& requiresSplitting);

	/** Main method to compute a simplex
	 *
	 * @param initialSet a polyhedra containing the initial set
	 * @param error min containment circle radius (linked to the error bound)
	 * @return polyhedra representing the simplex (null if  the initial set is too big)
	 */
	Simplex createSimplex(Hpolyhedron initialSet, Number error);

	/**
	 * Compute the maximal curvature by computing an important part of the transformation matrix
	 *
	 * @return the maximal curvature computed, if negative the computation has failed
	 */
	Number computeMaximalCurvature();

	/** Compute the transformation matrix
	 *
	 * @return: isotropic/anisotropic matrix transformation
	 */
	Matrix computeTransformationMatrix();

	/*
	 * ##########################################
	 * GETTTERS and SETTERS
	 * ##########################################
	 */

	Matrix getTransformation() const;
	void setTransformation(Matrix transformation);

	bool isUseTransformation() const;
	void setUseTransformation(bool useTransformation);

	double getMaxCurvature() {
		return maxCurvature;
	}
	void setMaxCurvature(double maxCurvature);

	void setUseDisplacement(bool useDisplacement) {
		this->useDisplacement = useDisplacement;
	}

	Box getCurvatureEstimationDomain() const {
		return curvatureEstimationDomain;
	}

	void setCurvatureEstimationDomain(Box curvatureEstimationDomain) {
		this->curvatureEstimationDomain = curvatureEstimationDomain;
	}

	unsigned getDimension() {
		return dimension;
	}

	void setNonlinearSystem(DynamicalSystem* nlsys) {
		if (nonlinearSystem != NULL)
			delete &nonlinearSystem;
		nonlinearSystem = nlsys;
		cout << "dim: " << nonlinearSystem->getDimension() << endl;
		nonlinearSystem->printType();
		nlsys->printType();
	}

private:

	/*
	 * ##########################################
	 * TRANSFORMATION MATRIX COMPUTATION
	 * ##########################################
	 */

	/** Compute the transformation matrix.
	 *
	 * @param curvature the curvature matrix associated to the system
	 * @return isotropic/anisotropic matrix transformation
	 */
	Matrix computeTransformationMatrix(Matrix &curvature);

	/** Compute the transformation matrix
	 *
	 * @param hessian list of system hessian matrix
	 * @return  isotropic/anisotropic matrix transformation
	 */
	Matrix computeTransformationMatrix(vector<Matrix>& hessian);

	/** Compute the Hessian matrix of the system
	 *
	 * use function derivative methods
	 * @return: list of system hessian matrix
	 */
	vector<Matrix> computeHessianMatrix();

	/** Compute the curvature matrix
	 *
	 * @param hessian a hessian matrix list
	 * @return curvature matrix computed from a list of hessian matrix
	 */
	Matrix computeCurvatureMatrix(vector<Matrix>& hessian);

	/*
	 * ##########################################
	 * SIMPLEX CREATION
	 * ##########################################
	 */

	/** create an initial simplex
	 *
	 * @param radius min containment circle radiusstdMatrix simplexVertices
	 * @param dimension the number of dimension of the system
	 * @return a regular simplex
	 */
	Simplex createRegularSimplex(Number radius, int dimension);

	/** create an initial simplex
	 *
	 * @param points a n simplex set of vertices
	 * @param direction a vector representing the system dynamic
	 */
	void alignToDirection(Vertices &points, Vector direction);

	/** place the simplex
	 *
	 * @param points a simplex represented by a list of vertices
	 * @param target a point
	 */
	bool placeSimplex(Vertices& points, Hpolyhedron& init,
			Number setMaxSize = 1);

	/** Compute a direction vector representing the dynamic of the system
	 *
	 * @param initial the current reachable set
	 * @return vector representing the system dynamic
	 */
	Vector computeDirectionalVector(Hpolyhedron &initial);

	/** Compute a polyhedron centroid
	 *
	 * @param poly a polyhedron
	 * @return point representing the centroid
	 */
	Vector computeCentroid(Hpolyhedron &poly);

	/** Compute a polyhedron centroid
	 *
	 * @param points a set of vertices
	 * @return point representing the centroid
	 */
	Vector computeCentroid(Vertices &points);

	/**
	 * Compute a box domain where the curvature will be estimated
	 * @param set the initial set or the las reachable set computed
	 */
	void computeCurvatureDomain(Hpolyhedron set);

	/**
	 * Test the curvature tensor
	 * @param curvature the curvature tensor
	 * @param domain domain to test the curvature
	 */
	void testCurvature(Matrix& curvature, int nbTest = 100);

	// use or not the isotropic transformation to increase the approximation domain volume
	bool useTransformation;

	unsigned dimension; // dimension of the system
	int type; // information on the system
	const Number minCurvature;

	vector<Matrix> cstHessian; //constant hessian matrix list
	vector<vector<DynamicalSystem*> > hessian; //not constant hessian matrix
	Matrix curvature; //curvature matrix
	Matrix transformation; //transformation matrix

	Simplex initialSimplex; //initial equilateral simplex
	bool useDisplacement; // use additional displacement when positionning the
						  //approximation domain around the initial set

	Number maxCurvature; // maximal curvature of the nonlinear system
	//TODO: for now we recompute the transformation matrix at each
	// 		approximation computation for non quadratic system,
	//		this should be optimized

	Box curvatureEstimationDomain; // domain where the curvature tensor is computed
	Number curvatureEstimationCoef; // a coefficient to determine the size of the curvature estimation domain

	DynamicalSystem* nonlinearSystem;

};

// ###############################
// Error Approximation computation
// ###############################

struct ErrorEstimationParameters {
	DynamicalSystem* originalSystem;
	DynamicalSystem* approximateSystem;
	Hpolyhedron domain;
	int indice;
};

Box estimateApproximationerror(DynamicalSystem* original,
		DynamicalSystem* approximation, Hpolyhedron& domain, Number initialerror);

}
 // namespace nltool

#endif

