/*
 * Box.h
 *
 *  Created on: Nov 20, 2012
 *      Author: romain
 */

#ifndef BOX_H_
#define BOX_H_

#include "Set.h"
#include "utils.h"
//#include "Hpolyhedron.h"
#include "PolynomialSystem.h"

namespace nltool {

class Hpolyhedron;

class Box: public Set {
public:

	/**
	 * Non-consistent constructor
	 */
	Box();

	/**
	 * Creates a box represented by lower and upper bounds
	 * @param l a vector corresponding to the lower bounds
	 * @param u a vector corresponding to the upper bounds
	 */
	Box(const stdVector& l, const stdVector& u);

	/**
	 * Creates the minimal box representation which contains a
	 * vertices list
	 * @param dim the set dimension
	 * @param vertices a list of vertices
	 */
	Box(unsigned dim, const Vertices& vertices);

	/**
	 * Creates a box that approximates a set of constraints given
	 * by the inequalities A*x<=b
	 * @param dim the set dimension
	 * @param A a matrix representing the left part of the inequalities
	 * @param b a vector representing the right part of the inequalities
	 */
	Box(unsigned dim, const Template& A, const Vector& b);

	/**
	 * Creates a box from a polyhedron represented by a set of constraints
	 * @param poly a polyhedron represented by a set of constrains
	 */
	Box(Hpolyhedron& poly);

	/**
	 * Destructor
	 */
	virtual ~Box();

	/**
	 * Performs a quick linear optimisation in this box
	 * @param direction optimization
	 * @param optPoint the problem solution
	 * @param max is a maximization problem (otherwise minimization)
	 * @return optimization result
	 */
	Number optimize(const Vector & direction, Vector & optPoint,
			bool max = true);

	/**
	 * Generates a list of 2^n vertices
	 * @return the vertices list
	 */
	Vertices generateVertices();

	/**
	 * Computes the box center
	 * @return the box center
	 */
	Vector centroid();

	/**
	 * Computes a set of affine functions that maps each interval of this
	 * box to the unit box (needed by the Bernstein based methods)
	 * @return a vector of affine functions
	 */
	std::vector<Polynom> computeUnitBoxMap();

	/**
	 * Inclusion test with poluhedron
	 * @param poly a polyhedron represented by a set of constraints
	 * @return is poly included in this box?
	 */
	bool include(Hpolyhedron& poly);

	/**
	 * Returns the maximal lenght of the box
	 * @return maximal lenghts
	 */
	Number getMaxInterval();

	/**
	 * Returns the minimal lenght of the box
	 * @return minimal lenghts
	 */
	Number getMinInterval();

	/*
	 * ##########################################
	 * GETTTERS and SETTERS
	 * ##########################################
	 */

	stdVector getBl() const;
	stdVector getBu() const;
	void setBl(stdVector bl);
	void setBu(stdVector bu);

private:
	stdVector bl;
	stdVector bu;
};

inline std::ostream& operator<<(std::ostream& out, Box box) {
	stdVector bl = box.getBl();
	stdVector bu = box.getBu();
	for (unsigned i = 0; i < box.getDimension(); i++) {
		out << "[" << bl[i] << ", " << bu[i] << "]";
		if (i < (box.getDimension() - 1))
			out << " * ";
	}
	out << std::endl;

	return out;
}

} // namespace nltool

#endif /* BOX_H_ */
