/*
 * Hpolyhedron.h
 *
 *  Created on: Nov 20, 2012
 *      Author: romain
 */

#ifndef HPOLYHEDRON_H_
#define HPOLYHEDRON_H_

#include "Set.h"
#include "utils.h"
#include "Box.h"
#include "Simplex.h"

namespace nltool {

class Hpolyhedron: public Set {
public:
	/**
	 * Non consistent constructor
	 */
	Hpolyhedron();

	/**
	 * Constructor
	 * @param dim dimension
	 * @param mat template matrix corresponding to the inequalities set
	 * (mat x <= vec)
	 * @param vec polyhedral coefficient corresponding to the inequalities set
	 * (mat x <= vec)
	 */
	Hpolyhedron(int dim, Template& mat, Vector& vec);

	Hpolyhedron(int dim, Template& mat, Vector& vec, Box& box);

	/**
	 * Constructor from simplex
	 * @param simplex simplex represented with vertices
	 */
	Hpolyhedron(Simplex& simplex);

	/**
	 * Constructor from box
	 * @param box a box represented by interval products
	 */
	Hpolyhedron(Box& box);

	/**
	 * Destructor
	 */
	virtual ~Hpolyhedron();

	/**
	 * Inclusion test
	 * @param poly another polyhedron
	 * @return return true if poly is included in this polyhedron, otherwise false
	 */
	bool includes(Hpolyhedron& poly);

	/**
	 * Optimization method (not implemented)
	 */
	Number optimize(const Vector & direction, Vector & optPoint,
			bool max = true) {
		return 0;
	}

	/**
	 * Inclusion test
	 * @param point a point
	 * @return return true if the point is included in this polyhedron, otherwise false
	 */
	bool includes(Vector& point);

	/**
	 * Splits automatically the polyhedron in two
	 * @return two polyhedron whose union is this polyhedron
	 */
	std::vector<Hpolyhedron> split();

	/**
	 * Cleans the set representation by removing redundant constraints
	 *  @return the number of redundant constraints removed
	 */
	int clean();

	/**
	 * returns a real number b coresponding to the position of a half-space
	 * with normal a defined by the inequality <a,x> <= b such that it supports
	 * the actual set
	 * @param a a directional vector
	 */
	Number support(Vector& a);

	/*
	 * ##########################################
	 * GETTTERS and SETTERS
	 * ##########################################
	 */

	Vector getCoefficients() const;
	Template getTemplates() const;

	void setCoefficients(Vector coefficients);
	void setTemplates(Template templates);

	Template getAllTemplates();
	Vector getAllCoefficients();

	Vertices getVertices();

	void inflate(Number inflat);

	const Box& getBbox() const {
		return bbox;
	}

	void setBbox(const Box& bbox) {
		this->bbox = bbox;
	}

private:
	Template templates;
	Vector coefficients;
	Box bbox; //Optional information
	Vertices vertices; //Optional information
};

inline std::ostream& operator<<(std::ostream& out, Hpolyhedron poly) {
	for (unsigned int i = 0; i < poly.getTemplates().size(); i++) {
		for (unsigned int j = 0; j < poly.getDimension() - 1; j++) {
			out << poly.getTemplates()[i][j] << "x" << j << " + ";
		}
		out << poly.getTemplates()[i][poly.getDimension() - 1] << "x"
				<< poly.getDimension() - 1 << " <= "
				<< poly.getCoefficients()[i] << std::endl;
	}
	out << std::endl;
	return out;
}

} // namespace nltool

#endif /* HPOLYHEDRON_H_ */
