/*
 * PointerSystem.h
 *
 *  Created on: Jan 22, 2013
 *      Author: romain
 */

#ifndef POINTERSYSTEM_H_
#define POINTERSYSTEM_H_

#include "DynamicalSystem.h"
#include "utils.h"
#include "Box.h"
#include "stdexcept"

namespace nltool {


class PointerSystem: public DynamicalSystem {
public:

	/**
	 * Non-consistent constructor
	 */
	PointerSystem();

	/**
	 * Creates a dynamical based on function pointer
	 * @param dim the system dimension
	 * @param f function to evaluate the dynamics at a given point
	 * @param df function derivated
	 * @param ddf function double derivated
	 * @param discretized is a discretized system?
	 */

	PointerSystem(int dim, Number (*f)(const int x, const stdVector &pt, void* data),
			Number (*df)(const int x, const stdVector &pt, const int dx, void* data),
			Number (*ddf)(const int x, const stdVector &pt, const int dx,
					const int dy, void* data), void* data, bool discretized = false);

	/**
	 * Destructor
	 */
	virtual ~PointerSystem();

	/**
	 * Evaluation function
	 * @param x the state of the system
	 * @return
	 */
	Vector eval(const Vector& x);

	/**
	 * Return the result of the double derivative of the system
	 * @param x a point in the state space
	 * @param dx first derivative variable indice in [0, dimension[
	 * @param dy second derivative variable indice in [0, dimension[
	 * @return
	 */
	Vector evalCurvature(const Vector& x, int dx, int dy);

	/**
	 * Compute the partial double derivative of the system
	 * @param i indice for the system row
	 * @param dx first derivative variable indice in [0, dimension[
	 * @param dy second derivative variable indice in [0, dimension[
	 * @param domain domain for the optimization
	 * @return
	 */
	Number getMaxCurvature(const unsigned i, const unsigned dx,
			const unsigned dy, const Box& domain);

	/**
	 * Optimization function proxy for the NL opt library
	 * @param x the initial guess in input and the optimization solution in output
	 * @param grad gradient
	 * @param f_data additional information
	 * @return
	 */
	static double CurvatureFunc(const std::vector<double> &x,
			std::vector<double> &grad, void* f_data);
	// variable used by CurvatureFunc
	unsigned curvatureFunctionX;
	unsigned curvatureFunctionDx;
	unsigned curvatureFunctionDy;
	void* dataPtr;
private:
	Number (*fPtr)(const int x, const stdVector &pt, void* data); //function studied f
	Number (*dfPtr)(const int x, const stdVector &pt, const int dx, void* data); // df/dx
	Number (*ddfPtr)(const int x, const stdVector &pt, const int dx,
			const int dy, void* data); // d^2f/dxdy


};

}  // namespace nltool


#endif /* POINTERSYSTEM_H_ */
