#include <iostream>
#include <string>
#include <fstream>
#include <iomanip>

//#include <stdio.h>
#include <stdlib.h>

#include <MatlabExporter.h>
#include <PolynomialSystem.h>
#include <utils.h>
#include <ApproximationBuilder.h>
#include <LinearSystem.h>
#include <ReachHybridization.h>
#include <Hpolyhedron.h>
#include <Box.h>
#include <Simplex.h>
#include <ReachPolynomial.h>
#include <HybridizationInterface.h>

#include <time.h>
#include <ObjExporter.h>
#include <test.h>

namespace nltool {

using namespace arma;

Hpolyhedron create2DSet(int nbfaces, double radius, doubleVector translation) {

	// initial constraint
	Template A(nbfaces, zeros(2));
	stdVector b(nbfaces, 0); // = arma::zeros(nbfaces);
	for (int i = 0; i < nbfaces; i++) {
		A[i][0] = cos(2.0 * i * PI / nbfaces);
		A[i][1] = sin(2.0 * i * PI / nbfaces);
		b[i] = radius + A[i][0] * translation[0] + A[i][1] * translation[1];
	}

	// bounding box
	stdVector bl(2);
	bl[0] = -radius + translation[0];
	bl[1] = -radius + translation[1];
	stdVector bu(2);
	bu[0] = radius + translation[0];
	bu[1] = radius + translation[1];
	Box box(bl, bu);
	Hpolyhedron result(box);

	return result;
}

PolynomialSystem test3D(Hpolyhedron& initialSet) {
	int dim = 3;

	std::string equations = "x0' = x0 - x1*x0 ";
	equations += "\n x1' = - x1 + x0*x1 - x1*x2";
	equations += "\n x2' = -x2 + x1*X2";
	PolynomialSystem result(dim, equations);

	// initial set
	stdVector bl(3, 0.49);
	bl[2] = 0.19;
	stdVector bu(3, 0.51);
	bu[2] = 0.21;

	doubleVector translation(dim, 0.5);
	translation[2] = 0.2;
	Box tmp(bl, bu);
	initialSet = Hpolyhedron(tmp);
	return result;
}

PolynomialSystem test2D(Hpolyhedron& initialSet) {

	std::string equations = "x0' = -1*x0^4 + 2*x1*x0^2";
	equations += "\n x1' = x1^3 - x0^2*x1^3";

	doubleVector translation(2, 1);
	initialSet = create2DSet(4, 0.01, translation);

	PolynomialSystem result(2, equations);
	return result;
}

PolynomialSystem species2D(Hpolyhedron& initialSet) {

	//species 2D
//	std::string equations =
//			"x0'= 0.1*x0 - 0.01*x1*x0 \n x1'= 0.001*x0*x1 - 0.05*x1";
//	doubleVector translation(2, 49.5);
//	translation[1] = 14.5;
//	initialSet = create2DSet(4, 0.1, translation);

	// neuronal
	std::string equations = "x0' = x0 + - 1*x0 ^3 + -1* x1 + 0.875 \n";
	equations += "x1' = 0.08*x0 + 0.056 + -0.064*x1";
	doubleVector translation(2, 1);
	translation[1] = 2.5;

	initialSet = create2DSet(4, 0.005, translation);

	// test
//		std::string equations = "x0' = x0^2 + x1^2 + 0.875 \n";
//		equations += "x1' = 2*x0^2 + 0.056";
//		doubleVector translation(2, 1);
//		translation[1] = 2.5;
//		initialSet = create2DSet(4, 0.01, translation);

	PolynomialSystem result(2, equations);

	// initial set

	// neuronal

	//result = result.getDiscretizedSystem(0.05);

	return result;
}

void testHybridization() {
	cout << "########### Hybridization simplex test" << endl;

	int nbIter = 39;
	int dim = 2;
	double errorBound = 0.01;
	double discretizationTimeStep = 0.02;
	Hpolyhedron initialSet;
	PolynomialSystem testSystem = species2D(initialSet);
	Vector center = Box(initialSet).centroid();

	cout << "system : \n" << testSystem << endl;
	cout << "initial set: \n" << Box(initialSet) << endl;

	//wait();

	DynamicalSystem* sysPtr;
	sysPtr = new PolynomialSystem(testSystem);

	MatlabExporter exporter(testSystem.getDimension());

	stdVector setColor(3, 0.2);
	setColor[0] = 1;

	ApproximationBuilder approximationBuilder(dim, sysPtr, false);
	vector<Hpolyhedron> result;
	result.push_back(initialSet);
	vector<Hpolyhedron> domains;
	Hpolyhedron currentDomain;
	bool splitting = false;
	LinearSystem approximation = approximationBuilder.createApproximation(
			initialSet, currentDomain, errorBound, splitting);
	ReachLinear reachLinear(dim, approximation, discretizationTimeStep);
	domains.push_back(currentDomain);
	for (unsigned int i = 0; i < nbIter; i++) {
		cout << "############# " << i << endl;
		initialSet = reachLinear.reach(initialSet);
		result.push_back(initialSet);
		if (!currentDomain.includes(initialSet)) {
			approximation = approximationBuilder.createApproximation(initialSet,
					currentDomain, errorBound, splitting);
			if (splitting)
				throw(runtime_error("splitting required"));
			domains.push_back(currentDomain);
			reachLinear = ReachLinear(dim, approximation,
					discretizationTimeStep);
		}
	}

//	ReachHybridization reachHybrid(sysPtr, initialSet, errorBound,
//			discretizationTimeStep);
//
//	reachHybrid.reach(nbIter);

//	for(unsigned i=0; i<reachables.size(); i++)
//		if((i%1) == 0)
//			reachablesPrinted.push_back(reachables[i]);
	exporter.save(domains);
	exporter.save(result, setColor);

	exporter.drawTrajectory(*sysPtr, center, nbIter * discretizationTimeStep,
			discretizationTimeStep);
	exporter.findBounds(domains);

//	exporter.saveVectorField(testSystem);

}

void testPolynomialMethod() {
	cout << "########### polynomial method test" << endl;
//int dim = 2;
//double errorBound = 0.1;
	Hpolyhedron initialSet;
	PolynomialSystem testSystem = species2D(initialSet);
	testSystem = testSystem.getDiscretizedSystem(0.5);

	ReachPolynomial solver(testSystem);

	MatlabExporter exporter(2);
	exporter.save(initialSet);

	int nbIter = 30;

	for (int i = 0; i < nbIter; i++) {
		initialSet = solver.reach(initialSet);
		cout << initialSet << endl;
		exporter.save(initialSet);
	}

}
//
//void testLinearReach4D() {
//	int dim = 4;
//	string filename = "test_linear.m";
//	Number dt = 0.05;
//	int nbIter =600;
//	std::vector<int> projection(2,2);
//	projection[0] = 0;
//	projection[1] = 1;
//
//
//	stdVector color(3, 0);
//	color[1] = 1;
//
//	Matrix A(zeros(dim, dim));
//	A(0, 0) = -0.003;
//	A(0, 1) = 0.039;
//	A(0, 2) = 0;
//	A(0, 3) = -0.322;
//
//	A(1, 0) = -0.065;
//	A(1, 1) = -0.319;
//	A(1, 2) = 7.740;
//	A(1, 3) = 0;
//
//	A(2, 0) = 0.02;
//	A(2, 1) = -0.101;
//	A(2, 2) = -0.429;
//	A(2, 3) = 0;
//
//	A(3, 0) = 0;
//	A(3, 1) = 0;
//	A(3, 2) = 1;
//	A(3, 3) = 0;
//
//	cout << "matrix A: " << A << endl;
//
//	Number deltaC = 0;//13.3;
//	stdVector blInput(4, -0.01 * deltaC);
//	stdVector buInput(4, 0.01 * deltaC);
//	blInput[1] = -0.18 * deltaC;
//	blInput[2] = -1.16 * deltaC;
//	blInput[3] = 0;
//
//	buInput[1] = 0.18 * deltaC;
//	buInput[2] = 1.16 * deltaC;
//	buInput[3] = 0;
//	Box inputSet(blInput, buInput);
//
//	Vector b(zeros(dim));
//
//	stdVector initBl(dim,1.0);
//	stdVector initBu(dim,2.1);
//	initBl[2] = 2.9;
//	initBu[2] = 3.1;
//	Box initialSetTmp(initBl, initBu);
//
//	LinearSystem linearSys(dim, A, b, static_cast<Set*>(&inputSet));
//
//	cout << "Linear system: " << linearSys << endl;
//
//	Hpolyhedron initial(initialSetTmp);
//	std::vector<Hpolyhedron> result;
//	result.push_back(initial);
//
//	ReachLinear solver(dim, linearSys, dt);
//
//	//	cout<<"eval: "<<linearSys.eval(testPoint)<<endl;
//
//
//	for (unsigned i = 0; i < nbIter; i++) {
//		initial = solver.reach(initial);
//		cout << initial << endl;
//		result.push_back(initial);
//	}
//
//	stdVector exporterBl, exporterBu = stdVector(dim);
//
//	MatlabExporter exporter(dim, filename,exporterBl, exporterBu, projection);
//	Template approxTemplate = createTemplateApproximation(dim,10,projection[0], projection[1]);
//	exporter.setApproximationTemplate(approxTemplate);
//	exporter.findBounds(result);
//
//	exporter.save(result);
//	//exporter.saveVectorField(linearSys);
//	cout << "DONE!!" << endl;
//}

void testLinearReach() {
	int dim = 2;
	string filename = "test_linear.m";
	Number dt = 0.01;
	stdVector blInput(2, -0.01);
	stdVector buInput(2, 0.01);
	stdVector bl(2, -2);
	stdVector bu(2, 2);

	stdVector color(3, 0);
	color[1] = 1;

	Matrix A(zeros(2, 2));
	A(0, 0) = -2;
	A(1, 1) = 2;
	A(1, 0) = 2;
	A(0, 1) = -4;

	cout << "matrix A: " << A << endl;

	Vector b(zeros(2));
	b[0] = 0;
	b[1] = 0;
	stdVector translation(2, 0);
	translation[0] = -1;

	Box inputSet(blInput, buInput);

	LinearSystem linearSys(dim, A, b, static_cast<Set*>(&inputSet));

	cout << "Linear system: " << linearSys << endl;

	Hpolyhedron initial = create2DSet(4, 0.1, translation);
	std::vector<Hpolyhedron> result;
	result.push_back(initial);

	ReachLinear solver(dim, linearSys, dt);

	for (unsigned i = 0; i < 50; i++) {
		initial = solver.reach(initial);
		cout << initial << endl;
		result.push_back(initial);
		//initial = result;
	}
	MatlabExporter exporter(2, filename, bl, bu);
	exporter.findBounds(result);
	exporter.save(result);
//exporter.saveVectorField(linearSys);
	cout << "DONE!!!!" << endl;
}

void testSimplexGLPK() {
	stdVector translation(2, 3);
	Hpolyhedron set(create2DSet(4, 1, translation));

	cout << "set: " << set << endl;

	Vector obj(zeros(2));
	obj[0] = -1;
	obj[1] = -1;

	Vector solution;
	Number res = simplex(obj, set.getTemplates(), set.getCoefficients(),
			solution, true);

	cout << "result: " << res << endl;
	cout << "solution: " << solution.t() << endl;

}

double testFuncF(const int x, const vector<double> &pt, void* data) {
	switch (x) {
	case 0:
		return (-pow(pt[0], 4) + 2 * pt[0] * pow(pt[1], 2));
		break;
	case 1:
		return (pow(pt[1], 3) - pow(pt[0], 2) * pow(pt[1], 3));
		break;
	default:
		break;
	}
	return 0;
}

double testFuncDf(const int x, const vector<double> &pt, const int dx,
		void* data) {
	return 0;
}

double testFuncDdf(const int x, const vector<double> &pt, const int dx,
		const int dy, void* data) {
	switch (x) {
	case 0:
		if ((dx == 0) && (dy == 0))
			return -12 * pow(pt[0], 2) + 4 * pt[1];
		if (((dx == 0) && (dy == 1)) || ((dx == 1) && (dy == 0)))
			return 2 * pt[0] * pt[0];
		break;
	case 1:
		if ((dx == 0) && (dy == 0))
			return -2 * pow(pt[1], 3);
		if ((dx == 1) && (dy == 1))
			return 6 * pt[1] - 6 * pow(pt[0], 2) * pt[1];
		if (((dx == 0) && (dy == 1)) || ((dx == 1) && (dy == 0)))
			return -6 * pt[0] * pow(pt[1], 2);
		break;
	default:
		break;
	}
	return 0;
}

void create2DSet(int nbfaces, double radius, vector<double> translation,
		vector<vector<double> > &resA, vector<double> &resB) {

// initial constraint
	resA = vector<vector<double> >(nbfaces, vector<double>(2, 0));
	resB = vector<double>(nbfaces);
	for (int i = 0; i < nbfaces; i++) {
		resA[i][0] = cos(2.0 * i * PI / nbfaces);
		resA[i][1] = sin(2.0 * i * PI / nbfaces);
		resB[i] = radius + resA[i][0] * translation[0]
				+ resA[i][1] * translation[1];
	}
}


void testlib() {

	cout << "test hybridization interface" << endl; // prints
	int dim = 2;

	vector<vector<double> > initialA, approxA, domainA;
	vector<double> initialB, approxB, domainB;

	vector<double> translation(2, 0);
	create2DSet(4, 0.1, translation, initialA, initialB);

	nltool::HybridizationInterface hybInterface(2);

	std::string sysStr("x0' = x0 + - 1*x0 ^3 + -1* x1 + 0.875 "
			"x1' = 0.08*x0 + 0.056 + -0.064*x1");
	hybInterface.setSystemDynamic(dim, sysStr);
//hybInterface.setSystemDynamic(dim, testFuncF, testFuncDdf);

	hybInterface.setInitialSet(dim, 4, initialA, initialB);
	int res = hybInterface.computeApproximation(1, approxA, approxB, domainA,
			domainB);

	cout << "\n\n#####result:\n\n";
	cout << "approximation: " << endl << "A: " << endl;
	for (unsigned i = 0; i < 2; i++) {
		for (unsigned j = 0; j < 2; j++)
			cout << approxA[i][j] << " ";
		cout << endl;
	}

	cout << endl << "b: " << endl;
	for (unsigned i = 0; i < 2; i++)
		cout << approxB[i] << " ";
	cout << endl;

	cout << "Approximation domain: " << endl;
	for (unsigned i = 0; i < 3; i++) {
		for (unsigned j = 0; j < 2; j++)
			cout << domainA[i][j] << "x" << j << " + ";
		cout << "<= " << domainB[i] << endl;
	}
	cout << "done" << endl;
}

} // namespace nltool

