/*
 * utils.cpp
 *
 *  Created on: Jan 10, 2011
 *      Author: romain
 */

/*
 * utils.h
 *
 *  Created on:
 *      Author: romain
 */

#include <utils.h>
#include <iostream>
#include <string>
#include <fstream>
#include <iomanip>
#include <stdio.h>
#include <vector>
#include <stdlib.h>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <glpk.h>
#include <time.h>

namespace nltool {

using namespace std;

bool useBinomTable = false;
int** binomialCoefficientTable;

// ###############################
// stdMatrix operation
// ###############################

stdMatrix matrixMult(const stdMatrix m1, const stdMatrix m2) {
	// we consider that m1 vector are in line et m2 vector are in column
	if (m1[0].size() != m2[0].size()) {
		std::cout << "stdMatrix size mismatch!!" << std::endl << "m1: "
				<< m1[0].size() << ", m2: " << m2.size() << std::endl;
		return m1;
	} else {
		stdMatrix result(m1.size(), doubleVector(m2.size(), 0.0));
		for (unsigned int i = 0; i < m1.size(); ++i) {
			for (unsigned int j = 0; j < m2.size(); ++j) {
				for (unsigned int k = 0; k < m1[0].size(); ++k) {
					result[i][j] += m1[i][k] * m2[j][k];
				}
			}
		}

		return result;
	}
}

stdMatrix matrixAdd(const stdMatrix m1, const stdMatrix m2) {
	if (m1[0].size() != m2[0].size()) {
		std::cout << "stdMatrix size mismatch!!" << std::endl << "m1: "
				<< m1.size() << " " << m1[0].size() << std::endl;
		std::cout << ", m2: " << m2.size() << " " << m2[0].size() << std::endl;
		return m1;
	} else {
		stdMatrix result(m1.size(), doubleVector(m1[0].size()));
		for (unsigned int i = 0; i < m1.size(); ++i) {
			for (unsigned int j = 0; j < m1[i].size(); ++j) {
				result[i][j] = m1[i][j] + m2[i][j];
			}
		}

		return result;
	}
}

stdMatrix matrixTranspose(const stdMatrix m1) {
	stdMatrix result(m1[0].size(), doubleVector(m1.size()));
	for (unsigned int i = 0; i < m1.size(); ++i) {
		for (unsigned int j = 0; j < m1[i].size(); ++j) {
			result[j][i] = m1[i][j];
		}
	}
	return result;
}

stdVector setUnitVector(const Vector v) {
	Number n = arma::norm(v, "inf");
	stdVector result(v.size());
	for (unsigned int i = 0; i < v.size(); i++) {
		result[i] = v[i] / n;
	}
	return result;
}

//extern "C" void dgchbv_(int*, double*, double[], int*, double[], double[],
//		int[], int*);
//extern "C" void dgpadm_(int*, int*, double*, double[], int*, double[], int*,
//		int[], int*, int*, int*);

// ###############################
// LP solver
// ###############################

inline double num2double(Number x) {
#ifdef GMP_RAT
	return x.get_d();
#else
	return x;
#endif
}

Number simplex(stdVector obj, stdMatrix Aineq, stdVector Bineq, stdMatrix Aeq,
		stdVector Beq, stdVector bl, stdVector bu, int &err) {
	Vector newObj = arma::conv_to<Vector>::from(obj);

	Template newTemplate;
	stdVector tmpVec;

	// copy inequalities
	for (unsigned i = 0; i < Aineq.size(); i++) {
		newTemplate.push_back(arma::conv_to<Vector>::from(Aineq[i]));
		tmpVec.push_back(Bineq[i]);
	}

//	// copy equalities
	for (unsigned i = 0; i < Aeq.size(); i++) {
		newTemplate.push_back(arma::conv_to<Vector>::from(Aeq[i]));
		tmpVec.push_back(Beq[i]);
		newTemplate.push_back(-(arma::conv_to<Vector>::from(Aeq[i])));
		tmpVec.push_back(-Beq[i]);
	}

	// copy bounds
	for (unsigned i = 0; i < bl.size(); i++) {
		Vector tmp = arma::zeros(bl.size());
		tmp[i] = 1;
		newTemplate.push_back(tmp);
		tmpVec.push_back(bu[i]);
		tmp[i] = -1;
		newTemplate.push_back(tmp);
		tmpVec.push_back(bl[i]);
	}

	Vector newCoef = arma::conv_to<Vector>::from(tmpVec);

	return simplex(newObj, newTemplate, newCoef, true);
}

Number simplex(Vector obj, Template M, Vector N, Vector& solution, bool max) {
	glp_prob *root = glp_create_prob();
	unsigned dim = obj.n_elem;
	double row[dim + 1];
	row[0] = 0;
	int ind[dim + 1];
	for (unsigned i = 0; i <= dim; i++)
		ind[i] = i;

	if (max)
		glp_set_obj_dir(root, GLP_MAX);
	else
		glp_set_obj_dir(root, GLP_MIN);

	glp_smcp param;
	glp_init_smcp(&param);
	param.msg_lev = GLP_MSG_ERR;

	glp_add_rows(root, M.size());
	glp_add_cols(root, dim);

	// copy the constraint
	for (unsigned i = 0; i < M.size(); i++) {
		// initialize the row
		for (unsigned j = 0; j < dim; j++)
			row[j + 1] = M[i][j];

		glp_set_mat_row(root, i + 1, dim, ind, row);
		glp_set_row_bnds(root, i + 1, GLP_UP, 0.0, N[i]);
	}

	// set objective function and bounds
	for (unsigned i = 0; i < dim; i++) {
		glp_set_obj_coef(root, i + 1, obj[i]);
		glp_set_col_bnds(root, i + 1, GLP_FR, 0, 0);
	}

	// solve the problem
	int info = glp_simplex(root, &param);

	if (info != 0) {
		glp_write_lp(root, NULL, "glpk_error.txt");
		glp_delete_prob(root);
		//glp_free_env();

		throw(runtime_error("Error in GLPK: " + info));
	}

	// retrieve results
	Number result = glp_get_obj_val(root);
	solution = arma::zeros(dim);
	for (unsigned i = 0; i < dim; i++)
		solution[i] = glp_get_col_prim(root, i + 1);

	// house keeping
	glp_delete_prob(root);
	//glp_free_env();

	return result;

}

Number simplex(Vector obj, Template M, Vector N, bool max) {
	Vector tmp;



	return simplex(obj, M, N, tmp, max);
}
//
//Number simplex(stdVector obj, Template M, stdVector N, stdVector bl,
//		stdVector bu, int &err) {
//	Template newTemplate =
//
//}
//
//	int lpdim = obj.size();
//	REAL row[lpdim + 1];
//	REAL row_obj[lpdim + 1];
//	lprec *lp;
//	lp = make_lp(0, lpdim);
//	set_verbose(lp, IMPORTANT);
//	set_add_rowmode(lp, TRUE);
//
//	//---- ADDING CONSTRAINTS
//	//* ATTENTION the program counts row/columns from 1..*/
//	for (unsigned int i = 0; i < M.size(); i++) {
//		row[0] = 0.0;
//		for (int j = 1; j <= lpdim; j++)
//			row[j] = num2double(M[i][j - 1]);
//		add_constraint(lp, row, LE, num2double(N[i]));
//	}
//
//	//---- ADDING BOUNDS
//	for (int j = 1; j <= lpdim; j++) {
//		set_lowbo(lp, j, num2double(bl[j - 1]));
//		set_upbo(lp, j, num2double(bu[j - 1]));
//	}
//
//	//---- ADDING OBJECTIVE FUNCTION
//	row_obj[0] = 0.0;
//	for (int j = 1; j <= lpdim; j++)
//		row_obj[j] = num2double(obj[j - 1]);
//	set_obj_fn(lp, row_obj);
//	set_minim(lp);
//	set_add_rowmode(lp, FALSE);
//
//	//---- SOLVING
//	int err_lp = solve(lp);
//	print_lp(lp);
//	if (err_lp != OPTIMAL && err_lp != SUBOPTIMAL) {
//		print_lp(lp);
//		throw(runtime_error("Error in LP solve: "+err_lp));
//		err = 1;
//		//return result;
//		return 0;
//	}
//	REAL result = get_objective(lp);
//	Number res = result;
//	delete_lp(lp);
//	return (res);
//}
//
//Number simplex(stdVector obj, stdMatrix Aineq, stdVector Bineq, stdMatrix Aeq,
//		stdVector Beq, stdVector bl, stdVector bu, int &err) {
//
//	int lpdim = Aineq[0].size();
//	REAL row[lpdim + 1];
//	REAL row_obj[lpdim + 1];
//	lprec *lp;
//	lp = make_lp(0, lpdim);
//	set_verbose(lp, IMPORTANT);
//	set_add_rowmode(lp, TRUE);
//
//	//---- ADDING CONSTRAINTS
//	//* ATTENTION the program counts row/columns from 1..*/
//	for (unsigned int i = 0; i < Aineq.size(); i++) {
//		row[0] = 0.0;
//		for (int j = 1; j <= lpdim; j++) {
//			row[j] = num2double(Aineq[i][j - 1]);
//		}
//		add_constraint(lp, row, LE, Bineq[i]);
//	}
//
//	for (unsigned int i = 0; i < Aeq.size(); i++) {
//		row[0] = 0.0;
//		for (int j = 1; j <= lpdim; j++) {
//			row[j] = num2double(row[j] = Aeq[i][j - 1]);
//		}
//		add_constraint(lp, row, EQ, Beq[i]);
//	}
//
//	//---- ADDING BOUNDS
//	for (int j = 1; j <= lpdim; j++) {
//		set_lowbo(lp, j, num2double(bl[j - 1]));
//		set_upbo(lp, j, num2double(bu[j - 1]));
//	}
//
//	//---- ADDING OBJECTIVE FUNCTION
//	row_obj[0] = 0.0;
//	for (int j = 1; j <= lpdim; j++)
//		row_obj[j] = num2double(obj[j - 1]);
//	set_obj_fn(lp, row_obj);
//	set_maxim(lp);
//	set_add_rowmode(lp, FALSE);
//
//	//---- SOLVING
//	int err_lp = solve(lp);
//	//print_lp(lp);
//	if (err_lp != OPTIMAL && err_lp != SUBOPTIMAL) {
//		err = 1;
//		//return result;
//		return 0;
//	}
//	//std::cout<<"system solved\n";
//	REAL result = get_objective(lp);
//
//	delete_lp(lp);
//	return (Number(result));
//
//}

/*
 *
 bool adjacencyTest(stdMatrix M, doubleVector N, unsigned int face1,
 unsigned int face2) {

 int lpdim = M[0].size();
 REAL row[lpdim + 1];
 REAL row_obj[lpdim + 1];
 lprec *lp;
 lp = make_lp(0, lpdim);
 set_verbose(lp, IMPORTANT);
 set_add_rowmode(lp, TRUE);

 //---- ADDING CONSTRAINTS
 //	// ATTENTION the program counts row/columns from 1..
 for (unsigned int i = 0; i < M.size(); i++) {
 row[0] = 0.0;
 for (int j = 1; j <= lpdim; j++)
 row[j] = M[i][j - 1];
 if (i == face1 || i == face2)
 add_constraint(lp, row, EQ, N[i]);
 else
 add_constraint(lp, row, LE, N[i]);
 }

 //---- ADDING BOUNDS
 for (int j = 1; j <= lpdim; j++) {
 set_lowbo(lp, j, -10e6);
 set_upbo(lp, j, 10e6);
 }

 //---- ADDING OBJECTIVE FUNCTION
 row_obj[0] = 0.0;
 for (int j = 1; j <= lpdim; j++)
 row_obj[j] = 1;../src/ReachPolynomial.cpp:418:76: error: no matching
 set_obj_fn(lp, row_obj);
 set_minim(lp);
 set_add_rowmode(lp, FALSE);

 //---- SOLVING
 int err_lp = solve(lp);
 //print_lp(lp);
 delete_lp(lp);
 if (err_lp != OPTIMAL && err_lp != SUBOPTIMAL) {
 //		cerr <<"LP solver failed\n";
 return false;
 }

 return true;
 }
 */

bool increaseMultiIndice(std::vector<int>& m, const std::vector<int> d) {
	unsigned int i = 0;
	while ((m[i] == d[i]) && (i < d.size())) {
		m[i] = 0;
		i++;
	}
	if (i < d.size()) {
		m[i] += 1;
		return true;
	} else
		return false;
}

// REDUNDANT
void createBinomialCoefficientTable(const int max) {
	useBinomTable = false;
	binomialCoefficientTable = (int **) malloc((max + 1) * sizeof(int));
	for (int i = 0; i <= max; i++)
		binomialCoefficientTable[i] = (int *) malloc((max + 1) * sizeof(int));
	for (int n = 0; n <= max; n++)
		for (int k = 0; k <= n; k++)
			binomialCoefficientTable[n][k] = binomialCoef(n, k);
	useBinomTable = true;
}

double binomialCoef(const int n, const int k) {
	if (useBinomTable) {
		//std::cout<<"binomial coeff used\n";
		return binomialCoefficientTable[n][k];
	}
	//std::cout<<"NO binomial coeff used\n";
	return fact(n) / (fact(k) * fact(n - k));
}

//
arma::mat expm_pad(const arma::mat &H, double t = 1.0, const int p = 6) {
	typedef unsigned size_type;
	typedef double real_value_type;

	assert(H.is_square());
	assert(p >= 1);
	const size_type n = H.n_rows;
	const arma::mat I = arma::eye(n, n);
	arma::mat U(n, n), H2(n, n), P(n, n), Q(n, n);
	real_value_type norm = 0.0;
// Calcuate Pade coefficients
	arma::colvec c(p + 1);
	c(0) = 1;
	for (size_type i = 0; i < p; ++i)
		c(i + 1) = c(i) * ((p - i) / ((i + 1.0) * (2.0 * p - i)));
// Calcuate the infinty norm of H, which is defined as the largest row sum of a matrix
	for (size_type i = 0; i < n; ++i) {
		real_value_type temp = 0.0;
		for (size_type j = 0; j < n; j++)
			temp += std::abs(H(i, j));
		if (norm < temp)
			norm = t * temp;
		else
			norm = t * norm;
		//norm = t * (std::max<real_value_type>(norm, temp));
	}
// If norm = 0, and all H elements are not NaN or infinity but zero,
// then U should be identity.
	if (norm == 0.0) {
		bool all_H_are_zero = true;
		for (size_type i = 0; i < n; i++)
			for (size_type j = 0; j < n; j++)
				if (H(i, j) != real_value_type(0.0))
					all_H_are_zero = false;
		if (all_H_are_zero == true)
			return I;
// Some error happens, H has elements which are NaN or infinity.
		std::cerr << "Null input error in the template expm_pad.\n";
		std::cout << "Null INPUT : " << H << "\n";
		exit(0);
	}
// Scaling, seek s such that || H*2^(-s) || < 1/2, and set scale = 2^(-s)
	int s = 0;
	double scale = 1.0;
	if (norm > 0.5) {
		s = std::max<int>(0, static_cast<int>((log(norm) / log(2.0) + 2.0)));
		scale /= real_value_type(std::pow(2.0, s));
		U = (scale * t) * H; // Here U is used as temp value due to that H is const
	} else
		U = H;

// Horner evaluation of the irreducible fraction, see the following ref above.
// Initialise P (numerator) and Q (denominator)
	H2 = U * U;
	Q = c(p) * I;
	P = c(p - 1) * I;
	size_type odd = 1;
	for (size_type k = p - 1; k > 0; --k) {
		(odd == 1) ? (Q = (Q * H2 + c(k - 1) * I)) : (P =
								(P * H2 + c(k - 1) * I));
		odd = 1 - odd;
	}
	(odd == 1) ? (Q = Q * U) : (P = P * U);
	Q -= P;

// Matrix Inversion
	H2 = arma::inv(Q);
	(odd == 1) ?
			(U = -(I + real_value_type(2.0) * (H2 * P))) :
			(U = I + real_value_type(2.0) * (H2 * P));
// Squaring
	for (size_type i = 0; i < s; ++i)
		U = U * U;
	return U;
}

Matrix matrixExponential(Matrix inA, Number dt) {
	return expm_pad(inA * dt);
}

Number simplify(const Number& num, const Number& precision) {
	return (ceil(num * (1.0 / precision)) * precision);
}

Vector simplify(const Vector& vec, const Number& precision) {
	Vector result(vec.size());
	for (unsigned i = 0; i < vec.size(); i++)
		result[i] = simplify(vec[i], precision);
	return result;
}


Template createTemplateApproximation(int dim, int nbfaces, int ind1, int ind2) {
	// initial constraint
	Template result(nbfaces, arma::zeros(dim));
	for (int i = 0; i < nbfaces; i++) {
		result[i][ind1] = cos(2.0 * i * PI / nbfaces);
		result[i][ind2] = sin(2.0 * i * PI / nbfaces);
	}
	return result;
}

void createOctagonalSet(int dim, double radius, Vector translation,
		Template &resA, Vector &resB) {
	Template templates;
	stdVector tmpCoef;
// initial constraint
	for (unsigned i = 0; i < dim; i++) {
		for (int k1 = -1; k1 < 2; k1++) {
			for (unsigned j = i; j < dim; j++) {
				for (int k2 = -1; k2 < 2; k2++) {
					if ((k1 != 0) && (k2 != 0)) {
						Vector normal = arma::zeros(dim);
						normal[i] = k1;
						normal[j] = k2;
						normal /= arma::norm(normal, 2);
						templates.push_back(normal);
						tmpCoef.push_back(
								radius + arma::dot(translation, normal));
					}
				}
			}
		}
	}

	resA = templates;
	resB = arma::conv_to<Vector>::from(tmpCoef);
}


Chrono::Chrono(): startClock(0){
}

Chrono::~Chrono() {
}

void Chrono::start() {
	startClock = clock();
}

Number Chrono::stop() {
	clock_t endClock = clock();

	return (endClock - startClock * 1.0)/CLOCKS_PER_SEC;
}

} // namespace nltool

//
//int main(int argc, char **argv) {
//	stdMatrix m(2,doubleVector(2));
//	m[0][0] = 1;
//	m[0][1] = 2;
//	m[1][0] = 3;
//	m[1][1] = 4;
//	int info;
//
//std::cout<<"stdMatrix :"<<m;
//	stdMatrix minv = inverse_matrix(2,m,info);
//std::cout<<"stdMatrix inverse: "<<minv;
//std::cout<<"test: "<<matrixMult(m,matrixTranspose(minv));
//	return 0;
//
//}


